import datetime
import json
from collections import defaultdict
from typing import Any
from typing import cast
from typing import DefaultDict
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Type

import apsw
from shillelagh.adapters.base import Adapter
from shillelagh.exceptions import ProgrammingError
from shillelagh.fields import Field
from shillelagh.fields import Order
from shillelagh.filters import Filter
from shillelagh.filters import Operator
from shillelagh.lib import deserialize
from shillelagh.types import Constraint
from shillelagh.types import Index
from shillelagh.types import RequestedOrder
from shillelagh.types import Row


operator_map = {
    apsw.SQLITE_INDEX_CONSTRAINT_EQ: Operator.EQ,
    apsw.SQLITE_INDEX_CONSTRAINT_GE: Operator.GE,
    apsw.SQLITE_INDEX_CONSTRAINT_GT: Operator.GT,
    apsw.SQLITE_INDEX_CONSTRAINT_LE: Operator.LE,
    apsw.SQLITE_INDEX_CONSTRAINT_LT: Operator.LT,
}


def convert_value(value: Any) -> Any:
    if isinstance(value, bool):
        return int(value)
    if isinstance(value, (int, float, str, type(None))):
        return value
    if isinstance(value, (datetime.datetime, datetime.date, datetime.time)):
        return value.isoformat()
    return str(value)


def convert_row(row: Row) -> Row:
    return {name: convert_value(value) for name, value in row.items()}


class VTModule:
    def __init__(self, adapter: Type[Adapter]):
        self.adapter = adapter

    def Create(
        self,
        connection: apsw.Connection,
        modulename: str,
        dbname: str,
        tablename: str,
        *args: str,
    ) -> Tuple[str, "VTTable"]:
        deserialized_args = [deserialize(arg) for arg in args]
        adapter = self.adapter(*deserialized_args)
        table = VTTable(adapter)
        create_table: str = table.get_create_table(tablename)
        return create_table, table

    Connect = Create


class VTTable:
    def __init__(self, adapter: Adapter):
        self.adapter = adapter

    def get_create_table(self, tablename: str) -> str:
        columns = self.adapter.get_columns()
        if not columns:
            raise ProgrammingError(f"Virtual table {tablename} has no columns")
        formatted_columns = ", ".join(f'"{k}" {v.type}' for (k, v) in columns.items())
        return f'CREATE TABLE "{tablename}" ({formatted_columns})'

    def BestIndex(
        self,
        constraints: List[Tuple[int, int]],
        orderbys: List[Tuple[int, bool]],
    ) -> Tuple[List[Constraint], int, str, bool, int]:
        column_types = list(self.adapter.get_columns().values())

        index_number = 42
        estimated_cost = 666

        indexes: List[Index] = []
        constraints_used: List[Constraint] = []
        filter_index = 0
        for column_index, sqlite_index_constraint in constraints:
            operator = operator_map.get(sqlite_index_constraint)
            column_type = column_types[column_index]
            for class_ in column_type.filters:
                if operator in class_.operators:
                    constraints_used.append((filter_index, column_type.exact))
                    filter_index += 1
                    indexes.append((column_index, sqlite_index_constraint))
                    break
            else:
                # no indexes supported in this column
                constraints_used.append(None)

        # is the data being returned in the requested order? if not, SQLite will have
        # to sort it
        orderby_consumed = True
        orderbys_to_process: List[Tuple[int, bool]] = []
        for column_index, descending in orderbys:
            requested_order = Order.DESCENDING if descending else Order.ASCENDING
            column_type = column_types[column_index]
            if column_type.order == Order.ANY:
                orderbys_to_process.append((column_index, descending))
            elif column_type.order != requested_order:
                orderby_consumed = False
                break

        # serialize the indexes to str so it can be used later when filtering the data
        index_name = json.dumps([indexes, orderbys_to_process])

        return (
            constraints_used,
            index_number,
            index_name,
            orderby_consumed,
            estimated_cost,
        )

    def Open(self) -> "VTCursor":
        return VTCursor(self.adapter)

    def Disconnect(self) -> None:
        self.adapter.close()

    Destroy = Disconnect

    def UpdateInsertRow(self, rowid: Optional[int], fields: Tuple[Any, ...]) -> int:
        column_names = ["rowid"] + list(self.adapter.get_columns().keys())
        row = dict(zip(column_names, [rowid] + list(fields)))
        return cast(int, self.adapter.insert_row(row))

    def UpdateDeleteRow(self, rowid: int) -> None:
        self.adapter.delete_row(rowid)

    def UpdateChangeRow(
        self,
        rowid: int,
        newrowid: int,
        fields: Tuple[Any, ...],
    ) -> None:
        column_names = ["rowid"] + list(self.adapter.get_columns().keys())
        row = dict(zip(column_names, [newrowid] + list(fields)))
        self.adapter.update_row(rowid, row)


class VTCursor:
    def __init__(self, adapter: Adapter):
        self.adapter = adapter

    def Filter(
        self,
        indexnumber: int,
        indexname: str,
        constraintargs: List[Any],
    ) -> None:
        columns: Dict[str, Field] = self.adapter.get_columns()
        column_names: List[str] = list(columns.keys())
        index = json.loads(indexname)
        indexes: List[Index] = index[0]
        orderbys: List[Tuple[int, bool]] = index[1]

        all_bounds: DefaultDict[str, Set[Tuple[Operator, Any]]] = defaultdict(set)
        for (column_index, sqlite_index_constraint), constraint in zip(
            indexes,
            constraintargs,
        ):
            if sqlite_index_constraint not in operator_map:
                raise Exception(f"Invalid constraint passed: {sqlite_index_constraint}")
            operator = operator_map[sqlite_index_constraint]
            column_name = column_names[column_index]
            column_type = columns[column_name]
            value = column_type.parse(constraint)
            all_bounds[column_name].add((operator, value))

        # find the filter that works with all the operations and build it
        bounds: Dict[str, Filter] = {}
        for column_name, operations in all_bounds.items():
            column_type = columns[column_name]
            operators = {operation[0] for operation in operations}
            for class_ in column_type.filters:
                if all(operator in class_.operators for operator in operators):
                    break
            else:
                raise Exception("No valid filter found")
            bounds[column_name] = class_.build(operations)

        order: List[Tuple[str, RequestedOrder]] = [
            (
                column_names[column_index],
                Order.DESCENDING if descending else Order.ASCENDING,
            )
            for column_index, descending in orderbys
        ]

        rows = (convert_row(row) for row in self.adapter.get_data(bounds, order))
        self.data = (
            tuple(row[name] for name in ["rowid"] + column_names) for row in rows
        )
        self.Next()

    def Eof(self) -> bool:
        return self.eof

    def Rowid(self) -> int:
        return cast(int, self.current_row[0])

    def Column(self, col) -> Any:
        return self.current_row[1 + col]

    def Next(self) -> None:
        try:
            self.current_row = next(self.data)
            self.eof = False
        except StopIteration:
            self.eof = True

    def Close(self) -> None:
        pass



"""Exceptions for DB API 2.0"""


class Warning(Exception):
    """
    Important warnings like data truncations while inserting.

    Exception raised for important warnings like data truncations
    while inserting, etc. It must be a subclass of the Python
    exceptions.StandardError (defined in the module exceptions).
    """


class Error(Exception):
    """
    Base class of all other error exceptions.

    Exception that is the base class of all other error exceptions.
    You can use this to catch all errors with one single except
    statement. Warnings are not considered errors and thus should
    not use this class as base. It must be a subclass of the Python
    StandardError (defined in the module exceptions).
    """


class InterfaceError(Error):
    """
    Errors that are related to the database interface.

    Exception raised for errors that are related to the database
    interface rather than the database itself. It must be a subclass
    of Error.
    """


class DatabaseError(Error):
    """
    Errors that are related to the database.

    Exception raised for errors that are related to the database.
    It must be a subclass of Error.
    """


class DataError(DatabaseError):
    """
    Errors that are due to problems with the processed data.

    Exception raised for errors that are due to problems with the
    processed data like division by zero, numeric value out of range,
    etc. It must be a subclass of DatabaseError.
    """


class OperationalError(DatabaseError):
    """
    Errors that are related to the database's operation.

    Exception raised for errors that are related to the database's
    operation and not necessarily under the control of the programmer,
    e.g. an unexpected disconnect occurs, the data source name is not
    found, a transaction could not be processed, a memory allocation
    error occurred during processing, etc. It must be a subclass of
    DatabaseError.
    """


class IntegrityError(DatabaseError):
    """
    Raised when the relational integrity of the database is affected.

    Exception raised when the relational integrity of the database is
    affected, e.g. a foreign key check fails. It must be a subclass of
    DatabaseError.
    """


class InternalError(DatabaseError):
    """
    Raised when the database encounters an internal error.

    Exception raised when the database encounters an internal error,
    e.g. the cursor is not valid anymore, the transaction is out of
    sync, etc. It must be a subclass of DatabaseError.
    """


class ProgrammingError(DatabaseError):
    """
    Raised for programming errors.

    Exception raised for programming errors, e.g. table not found or
    already exists, syntax error in the SQL statement, wrong number
    of parameters specified, etc. It must be a subclass of DatabaseError.
    """


class NotSupportedError(DatabaseError):
    """
    Raised in case a method or database API is not supported.

    Exception raised in case a method or database API was used which is
    not supported by the database, e.g. requesting a .rollback() on a
    connection that does not support transaction or has transactions
    turned off. It must be a subclass of DatabaseError.
    """



import operator
import urllib.parse
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

from shillelagh.adapters.base import Adapter
from shillelagh.fields import Float
from shillelagh.fields import Integer
from shillelagh.fields import Order
from shillelagh.fields import String
from shillelagh.filters import Equal
from shillelagh.filters import Filter
from shillelagh.filters import Range
from shillelagh.types import RequestedOrder
from shillelagh.types import Row


class FakeEntryPoint:
    def __init__(self, name: str, adapter: Adapter):
        self.name = name
        self.adapter = adapter

    def load(self) -> Adapter:
        return self.adapter


class FakeAdapter(Adapter):

    safe = False

    age = Float(filters=[Range], order=Order.ANY, exact=True)
    name = String(filters=[Equal], order=Order.ANY, exact=True)
    pets = Integer(order=Order.ANY)

    @staticmethod
    def supports(uri: str) -> bool:
        parsed = urllib.parse.urlparse(uri)
        return parsed.scheme == "dummy"

    @staticmethod
    def parse_uri(uri: str) -> Tuple[()]:
        return ()

    def __init__(self):
        self.data = [
            {"rowid": 0, "name": "Alice", "age": 20, "pets": 0},
            {"rowid": 1, "name": "Bob", "age": 23, "pets": 3},
        ]

    def get_data(
        self,
        bounds: Dict[str, Filter],
        order: List[Tuple[str, RequestedOrder]],
    ) -> Iterator[Dict[str, Any]]:
        data = self.data[:]

        for column in ["name", "age"]:
            if column in bounds:
                data = [row for row in data if bounds[column].check(row[column])]

        for column_name, requested_order in order:
            reverse = requested_order == Order.DESCENDING
            data.sort(key=operator.itemgetter(column_name), reverse=reverse)

        yield from iter(data)

    def insert_row(self, row: Row) -> int:
        row_id: Optional[int] = row["rowid"]
        if row_id is None:
            row["rowid"] = row_id = max(row["rowid"] for row in self.data) + 1

        self.data.append(row)

        return row_id

    def delete_row(self, row_id: int) -> None:
        self.data = [row for row in self.data if row["rowid"] != row_id]


weatherapi_response = {
    "location": {
        "name": "Reykjavik",
        "region": "Gullbringusysla",
        "country": "Iceland",
        "lat": 64.15,
        "lon": -21.95,
        "tz_id": "Atlantic/Reykjavik",
        "localtime_epoch": 1616202420,
        "localtime": "2021-03-20 1:07",
    },
    "forecast": {
        "forecastday": [
            {
                "date": "2021-03-17",
                "date_epoch": 1615939200,
                "day": {
                    "maxtemp_c": 7.4,
                    "maxtemp_f": 45.3,
                    "mintemp_c": 5.5,
                    "mintemp_f": 41.9,
                    "avgtemp_c": 6.7,
                    "avgtemp_f": 44.1,
                    "maxwind_mph": 16.3,
                    "maxwind_kph": 26.3,
                    "totalprecip_mm": 2.3,
                    "totalprecip_in": 0.09,
                    "avgvis_km": 7.8,
                    "avgvis_miles": 4.0,
                    "avghumidity": 95.0,
                    "condition": {
                        "text": "Light rain",
                        "icon": "//cdn.weatherapi.com/weather/64x64/day/296.png",
                        "code": 1183,
                    },
                    "uv": 0.0,
                },
                "astro": {
                    "sunrise": "07:39 AM",
                    "sunset": "07:36 PM",
                    "moonrise": "08:37 AM",
                    "moonset": "No moonset",
                    "moon_phase": "Waxing Crescent",
                    "moon_illumination": "23",
                },
                "hour": [
                    {
                        "time_epoch": 1615939200,
                        "time": "2021-03-17 00:00",
                        "temp_c": 5.5,
                        "temp_f": 41.9,
                        "is_day": 0,
                        "condition": {
                            "text": "Overcast",
                            "icon": "//cdn.weatherapi.com/weather/64x64/night/122.png",
                            "code": 1009,
                        },
                        "wind_mph": 13.2,
                        "wind_kph": 21.2,
                        "wind_degree": 134,
                        "wind_dir": "SE",
                        "pressure_mb": 1021.0,
                        "pressure_in": 30.6,
                        "precip_mm": 0.0,
                        "precip_in": 0.0,
                        "humidity": 95,
                        "cloud": 100,
                        "feelslike_c": 1.6,
                        "feelslike_f": 34.9,
                        "windchill_c": 1.6,
                        "windchill_f": 34.9,
                        "heatindex_c": 5.5,
                        "heatindex_f": 41.9,
                        "dewpoint_c": 4.7,
                        "dewpoint_f": 40.5,
                        "will_it_rain": 0,
                        "chance_of_rain": "0",
                        "will_it_snow": 0,
                        "chance_of_snow": "0",
                        "vis_km": 10.0,
                        "vis_miles": 6.0,
                        "gust_mph": 22.1,
                        "gust_kph": 35.6,
                    },
                    {
                        "time_epoch": 1615942800,
                        "time": "2021-03-17 01:00",
                        "temp_c": 5.5,
                        "temp_f": 41.9,
                        "is_day": 0,
                        "condition": {
                            "text": "Overcast",
                            "icon": "//cdn.weatherapi.com/weather/64x64/night/122.png",
                            "code": 1009,
                        },
                        "wind_mph": 12.8,
                        "wind_kph": 20.6,
                        "wind_degree": 133,
                        "wind_dir": "SE",
                        "pressure_mb": 1020.0,
                        "pressure_in": 30.6,
                        "precip_mm": 0.0,
                        "precip_in": 0.0,
                        "humidity": 95,
                        "cloud": 100,
                        "feelslike_c": 1.7,
                        "feelslike_f": 35.0,
                        "windchill_c": 1.7,
                        "windchill_f": 35.0,
                        "heatindex_c": 5.5,
                        "heatindex_f": 41.9,
                        "dewpoint_c": 4.8,
                        "dewpoint_f": 40.6,
                        "will_it_rain": 0,
                        "chance_of_rain": "3",
                        "will_it_snow": 0,
                        "chance_of_snow": "0",
                        "vis_km": 9.7,
                        "vis_miles": 6.0,
                        "gust_mph": 22.1,
                        "gust_kph": 35.5,
                    },
                    {
                        "time_epoch": 1615946400,
                        "time": "2021-03-17 02:00",
                        "temp_c": 5.5,
                        "temp_f": 41.9,
                        "is_day": 0,
                        "condition": {
                            "text": "Light rain",
                            "icon": "//cdn.weatherapi.com/weather/64x64/night/296.png",
                            "code": 1183,
                        },
                        "wind_mph": 12.5,
                        "wind_kph": 20.0,
                        "wind_degree": 131,
                        "wind_dir": "SE",
                        "pressure_mb": 1020.0,
                        "pressure_in": 30.6,
                        "precip_mm": 0.27,
                        "precip_in": 0.01,
                        "humidity": 96,
                        "cloud": 100,
                        "feelslike_c": 1.7,
                        "feelslike_f": 35.1,
                        "windchill_c": 1.7,
                        "windchill_f": 35.1,
                        "heatindex_c": 5.5,
                        "heatindex_f": 41.9,
                        "dewpoint_c": 4.8,
                        "dewpoint_f": 40.7,
                        "will_it_rain": 0,
                        "chance_of_rain": "61",
                        "will_it_snow": 0,
                        "chance_of_snow": "0",
                        "vis_km": 9.3,
                        "vis_miles": 5.0,
                        "gust_mph": 22.0,
                        "gust_kph": 35.4,
                    },
                    {
                        "time_epoch": 1615950000,
                        "time": "2021-03-17 03:00",
                        "temp_c": 5.5,
                        "temp_f": 41.9,
                        "is_day": 0,
                        "condition": {
                            "text": "Light rain",
                            "icon": "//cdn.weatherapi.com/weather/64x64/night/296.png",
                            "code": 1183,
                        },
                        "wind_mph": 12.1,
                        "wind_kph": 19.4,
                        "wind_degree": 130,
                        "wind_dir": "SE",
                        "pressure_mb": 1020.0,
                        "pressure_in": 30.6,
                        "precip_mm": 0.36,
                        "precip_in": 0.01,
                        "humidity": 96,
                        "cloud": 100,
                        "feelslike_c": 1.8,
                        "feelslike_f": 35.2,
                        "windchill_c": 1.8,
                        "windchill_f": 35.2,
                        "heatindex_c": 5.5,
                        "heatindex_f": 41.9,
                        "dewpoint_c": 4.9,
                        "dewpoint_f": 40.8,
                        "will_it_rain": 0,
                        "chance_of_rain": "0",
                        "will_it_snow": 0,
                        "chance_of_snow": "0",
                        "vis_km": 9.0,
                        "vis_miles": 5.0,
                        "gust_mph": 21.9,
                        "gust_kph": 35.3,
                    },
                    {
                        "time_epoch": 1615953600,
                        "time": "2021-03-17 04:00",
                        "temp_c": 5.6,
                        "temp_f": 42.1,
                        "is_day": 0,
                        "condition": {
                            "text": "Light rain",
                            "icon": "//cdn.weatherapi.com/weather/64x64/night/296.png",
                            "code": 1183,
                        },
                        "wind_mph": 11.7,
                        "wind_kph": 18.8,
                        "wind_degree": 133,
                        "wind_dir": "SE",
                        "pressure_mb": 1020.0,
                        "pressure_in": 30.6,
                        "precip_mm": 0.18,
                        "precip_in": 0.01,
                        "humidity": 96,
                        "cloud": 100,
                        "feelslike_c": 2.0,
                        "feelslike_f": 35.6,
                        "windchill_c": 2.0,
                        "windchill_f": 35.6,
                        "heatindex_c": 5.6,
                        "heatindex_f": 42.1,
                        "dewpoint_c": 5.0,
                        "dewpoint_f": 41.0,
                        "will_it_rain": 0,
                        "chance_of_rain": "61",
                        "will_it_snow": 0,
                        "chance_of_snow": "0",
                        "vis_km": 9.0,
                        "vis_miles": 5.0,
                        "gust_mph": 21.6,
                        "gust_kph": 34.8,
                    },
                    {
                        "time_epoch": 1615957200,
                        "time": "2021-03-17 05:00",
                        "temp_c": 5.7,
                        "temp_f": 42.3,
                        "is_day": 0,
                        "condition": {
                            "text": "Light rain",
                            "icon": "//cdn.weatherapi.com/weather/64x64/night/296.png",
                            "code": 1183,
                        },
                        "wind_mph": 11.3,
                        "wind_kph": 18.2,
                        "wind_degree": 135,
                        "wind_dir": "SE",
                        "pressure_mb": 1020.0,
                        "pressure_in": 30.6,
                        "precip_mm": 0.27,
                        "precip_in": 0.01,
                        "humidity": 96,
                        "cloud": 100,
                        "feelslike_c": 2.2,
                        "feelslike_f": 36.0,
                        "windchill_c": 2.2,
                        "windchill_f": 36.0,
                        "heatindex_c": 5.7,
                        "heatindex_f": 42.3,
                        "dewpoint_c": 5.1,
                        "dewpoint_f": 41.2,
                        "will_it_rain": 0,
                        "chance_of_rain": "61",
                        "will_it_snow": 0,
                        "chance_of_snow": "0",
                        "vis_km": 9.0,
                        "vis_miles": 5.0,
                        "gust_mph": 21.3,
                        "gust_kph": 34.3,
                    },
                    {
                        "time_epoch": 1615960800,
                        "time": "2021-03-17 06:00",
                        "temp_c": 5.8,
                        "temp_f": 42.4,
                        "is_day": 0,
                        "condition": {
                            "text": "Light rain",
                            "icon": "//cdn.weatherapi.com/weather/64x64/night/296.png",
                            "code": 1183,
                        },
                        "wind_mph": 11.0,
                        "wind_kph": 17.6,
                        "wind_degree": 138,
                        "wind_dir": "SE",
                        "pressure_mb": 1020.0,
                        "pressure_in": 30.6,
                        "precip_mm": 0.36,
                        "precip_in": 0.01,
                        "humidity": 96,
                        "cloud": 100,
                        "feelslike_c": 2.4,
                        "feelslike_f": 36.3,
                        "windchill_c": 2.4,
                        "windchill_f": 36.3,
                        "heatindex_c": 5.8,
                        "heatindex_f": 42.4,
                        "dewpoint_c": 5.2,
                        "dewpoint_f": 41.4,
                        "will_it_rain": 0,
                        "chance_of_rain": "0",
                        "will_it_snow": 0,
                        "chance_of_snow": "0",
                        "vis_km": 9.0,
                        "vis_miles": 5.0,
                        "gust_mph": 21.0,
                        "gust_kph": 33.8,
                    },
                    {
                        "time_epoch": 1615964400,
                        "time": "2021-03-17 07:00",
                        "temp_c": 6.0,
                        "temp_f": 42.9,
                        "is_day": 0,
                        "condition": {
                            "text": "Light rain",
                            "icon": "//cdn.weatherapi.com/weather/64x64/night/296.png",
                            "code": 1183,
                        },
                        "wind_mph": 11.3,
                        "wind_kph": 18.1,
                        "wind_degree": 146,
                        "wind_dir": "SSE",
                        "pressure_mb": 1021.0,
                        "pressure_in": 30.6,
                        "precip_mm": 0.18,
                        "precip_in": 0.01,
                        "humidity": 96,
                        "cloud": 100,
                        "feelslike_c": 2.6,
                        "feelslike_f": 36.7,
                        "windchill_c": 2.6,
                        "windchill_f": 36.7,
                        "heatindex_c": 6.0,
                        "heatindex_f": 42.9,
                        "dewpoint_c": 5.4,
                        "dewpoint_f": 41.8,
                        "will_it_rain": 0,
                        "chance_of_rain": "61",
                        "will_it_snow": 0,
                        "chance_of_snow": "0",
                        "vis_km": 6.7,
                        "vis_miles": 4.0,
                        "gust_mph": 21.5,
                        "gust_kph": 34.6,
                    },
                    {
                        "time_epoch": 1615968000,
                        "time": "2021-03-17 08:00",
                        "temp_c": 6.3,
                        "temp_f": 43.3,
                        "is_day": 1,
                        "condition": {
                            "text": "Light drizzle",
                            "icon": "//cdn.weatherapi.com/weather/64x64/day/266.png",
                            "code": 1153,
                        },
                        "wind_mph": 11.6,
                        "wind_kph": 18.6,
                        "wind_degree": 154,
                        "wind_dir": "SSE",
                        "pressure_mb": 1021.0,
                        "pressure_in": 30.6,
                        "precip_mm": 0.1,
                        "precip_in": 0.0,
                        "humidity": 96,
                        "cloud": 100,
                        "feelslike_c": 2.9,
                        "feelslike_f": 37.2,
                        "windchill_c": 2.9,
                        "windchill_f": 37.2,
                        "heatindex_c": 6.3,
                        "heatindex_f": 43.3,
                        "dewpoint_c": 5.7,
                        "dewpoint_f": 42.2,
                        "will_it_rain": 0,
                        "chance_of_rain": "51",
                        "will_it_snow": 0,
                        "chance_of_snow": "0",
                        "vis_km": 4.3,
                        "vis_miles": 2.0,
                        "gust_mph": 21.9,
                        "gust_kph": 35.3,
                    },
                    {
                        "time_epoch": 1615971600,
                        "time": "2021-03-17 09:00",
                        "temp_c": 6.5,
                        "temp_f": 43.7,
                        "is_day": 1,
                        "condition": {
                            "text": "Light drizzle",
                            "icon": "//cdn.weatherapi.com/weather/64x64/day/266.png",
                            "code": 1153,
                        },
                        "wind_mph": 11.9,
                        "wind_kph": 19.1,
                        "wind_degree": 162,
                        "wind_dir": "SSE",
                        "pressure_mb": 1022.0,
                        "pressure_in": 30.7,
                        "precip_mm": 0.13,
                        "precip_in": 0.01,
                        "humidity": 96,
                        "cloud": 100,
                        "feelslike_c": 3.1,
                        "feelslike_f": 37.6,
                        "windchill_c": 3.1,
                        "windchill_f": 37.6,
                        "heatindex_c": 6.5,
                        "heatindex_f": 43.7,
                        "dewpoint_c": 5.9,
                        "dewpoint_f": 42.6,
                        "will_it_rain": 0,
                        "chance_of_rain": "0",
                        "will_it_snow": 0,
                        "chance_of_snow": "0",
                        "vis_km": 2.0,
                        "vis_miles": 1.0,
                        "gust_mph": 22.4,
                        "gust_kph": 36.0,
                    },
                    {
                        "time_epoch": 1615975200,
                        "time": "2021-03-17 10:00",
                        "temp_c": 6.6,
                        "temp_f": 43.8,
                        "is_day": 1,
                        "condition": {
                            "text": "Light drizzle",
                            "icon": "//cdn.weatherapi.com/weather/64x64/day/266.png",
                            "code": 1153,
                        },
                        "wind_mph": 11.9,
                        "wind_kph": 19.2,
                        "wind_degree": 154,
                        "wind_dir": "SSE",
                        "pressure_mb": 1022.0,
                        "pressure_in": 30.7,
                        "precip_mm": 0.07,
                        "precip_in": 0.0,
                        "humidity": 96,
                        "cloud": 97,
                        "feelslike_c": 3.2,
                        "feelslike_f": 37.7,
                        "windchill_c": 3.2,
                        "windchill_f": 37.7,
                        "heatindex_c": 6.6,
                        "heatindex_f": 43.8,
                        "dewpoint_c": 6.0,
                        "dewpoint_f": 42.7,
                        "will_it_rain": 0,
                        "chance_of_rain": "51",
                        "will_it_snow": 0,
                        "chance_of_snow": "0",
                        "vis_km": 2.0,
                        "vis_miles": 1.0,
                        "gust_mph": 22.1,
                        "gust_kph": 35.5,
                    },
                    {
                        "time_epoch": 1615978800,
                        "time": "2021-03-17 11:00",
                        "temp_c": 6.6,
                        "temp_f": 43.9,
                        "is_day": 1,
                        "condition": {
                            "text": "Light drizzle",
                            "icon": "//cdn.weatherapi.com/weather/64x64/day/266.png",
                            "code": 1153,
                        },
                        "wind_mph": 12.0,
                        "wind_kph": 19.3,
                        "wind_degree": 146,
                        "wind_dir": "SSE",
                        "pressure_mb": 1023.0,
                        "pressure_in": 30.7,
                        "precip_mm": 0.1,
                        "precip_in": 0.0,
                        "humidity": 96,
                        "cloud": 93,
                        "feelslike_c": 3.2,
                        "feelslike_f": 37.8,
                        "windchill_c": 3.2,
                        "windchill_f": 37.8,
                        "heatindex_c": 6.6,
                        "heatindex_f": 43.9,
                        "dewpoint_c": 6.0,
                        "dewpoint_f": 42.9,
                        "will_it_rain": 0,
                        "chance_of_rain": "51",
                        "will_it_snow": 0,
                        "chance_of_snow": "0",
                        "vis_km": 2.0,
                        "vis_miles": 1.0,
                        "gust_mph": 21.8,
                        "gust_kph": 35.0,
                    },
                    {
                        "time_epoch": 1615982400,
                        "time": "2021-03-17 12:00",
                        "temp_c": 6.7,
                        "temp_f": 44.1,
                        "is_day": 1,
                        "condition": {
                            "text": "Light drizzle",
                            "icon": "//cdn.weatherapi.com/weather/64x64/day/266.png",
                            "code": 1153,
                        },
                        "wind_mph": 12.1,
                        "wind_kph": 19.4,
                        "wind_degree": 139,
                        "wind_dir": "SE",
                        "pressure_mb": 1023.0,
                        "pressure_in": 30.7,
                        "precip_mm": 0.13,
                        "precip_in": 0.01,
                        "humidity": 96,
                        "cloud": 90,
                        "feelslike_c": 3.3,
                        "feelslike_f": 37.9,
                        "windchill_c": 3.3,
                        "windchill_f": 37.9,
                        "heatindex_c": 6.7,
                        "heatindex_f": 44.1,
                        "dewpoint_c": 6.1,
                        "dewpoint_f": 43.0,
                        "will_it_rain": 0,
                        "chance_of_rain": "0",
                        "will_it_snow": 0,
                        "chance_of_snow": "0",
                        "vis_km": 2.0,
                        "vis_miles": 1.0,
                        "gust_mph": 21.5,
                        "gust_kph": 34.6,
                    },
                    {
                        "time_epoch": 1615986000,
                        "time": "2021-03-17 13:00",
                        "temp_c": 6.9,
                        "temp_f": 44.5,
                        "is_day": 1,
                        "condition": {
                            "text": "Light drizzle",
                            "icon": "//cdn.weatherapi.com/weather/64x64/day/266.png",
                            "code": 1153,
                        },
                        "wind_mph": 12.9,
                        "wind_kph": 20.8,
                        "wind_degree": 137,
                        "wind_dir": "SE",
                        "pressure_mb": 1023.0,
                        "pressure_in": 30.7,
                        "precip_mm": 0.07,
                        "precip_in": 0.0,
                        "humidity": 95,
                        "cloud": 85,
                        "feelslike_c": 3.5,
                        "feelslike_f": 38.2,
                        "windchill_c": 3.5,
                        "windchill_f": 38.2,
                        "heatindex_c": 6.9,
                        "heatindex_f": 44.5,
                        "dewpoint_c": 6.2,
                        "dewpoint_f": 43.2,
                        "will_it_rain": 0,
                        "chance_of_rain": "51",
                        "will_it_snow": 0,
                        "chance_of_snow": "0",
                        "vis_km": 4.7,
                        "vis_miles": 2.0,
                        "gust_mph": 22.8,
                        "gust_kph": 36.7,
                    },
                    {
                        "time_epoch": 1615989600,
                        "time": "2021-03-17 14:00",
                        "temp_c": 7.2,
                        "temp_f": 44.9,
                        "is_day": 1,
                        "condition": {
                            "text": "Cloudy",
                            "icon": "//cdn.weatherapi.com/weather/64x64/day/119.png",
                            "code": 1006,
                        },
                        "wind_mph": 13.7,
                        "wind_kph": 22.1,
                        "wind_degree": 136,
                        "wind_dir": "SE",
                        "pressure_mb": 1024.0,
                        "pressure_in": 30.7,
                        "precip_mm": 0.0,
                        "precip_in": 0.0,
                        "humidity": 94,
                        "cloud": 81,
                        "feelslike_c": 3.6,
                        "feelslike_f": 38.5,
                        "windchill_c": 3.6,
                        "windchill_f": 38.5,
                        "heatindex_c": 7.2,
                        "heatindex_f": 44.9,
                        "dewpoint_c": 6.3,
                        "dewpoint_f": 43.3,
                        "will_it_rain": 0,
                        "chance_of_rain": "2",
                        "will_it_snow": 0,
                        "chance_of_snow": "0",
                        "vis_km": 7.3,
                        "vis_miles": 4.0,
                        "gust_mph": 24.2,
                        "gust_kph": 38.9,
                    },
                    {
                        "time_epoch": 1615993200,
                        "time": "2021-03-17 15:00",
                        "temp_c": 7.4,
                        "temp_f": 45.3,
                        "is_day": 1,
                        "condition": {
                            "text": "Cloudy",
                            "icon": "//cdn.weatherapi.com/weather/64x64/day/119.png",
                            "code": 1006,
                        },
                        "wind_mph": 14.5,
                        "wind_kph": 23.4,
                        "wind_degree": 135,
                        "wind_dir": "SE",
                        "pressure_mb": 1024.0,
                        "pressure_in": 30.7,
                        "precip_mm": 0.0,
                        "precip_in": 0.0,
                        "humidity": 93,
                        "cloud": 76,
                        "feelslike_c": 3.8,
                        "feelslike_f": 38.8,
                        "windchill_c": 3.8,
                        "windchill_f": 38.8,
                        "heatindex_c": 7.4,
                        "heatindex_f": 45.3,
                        "dewpoint_c": 6.4,
                        "dewpoint_f": 43.5,
                        "will_it_rain": 0,
                        "chance_of_rain": "0",
                        "will_it_snow": 0,
                        "chance_of_snow": "0",
                        "vis_km": 10.0,
                        "vis_miles": 6.0,
                        "gust_mph": 25.5,
                        "gust_kph": 41.0,
                    },
                    {
                        "time_epoch": 1615996800,
                        "time": "2021-03-17 16:00",
                        "temp_c": 7.3,
                        "temp_f": 45.1,
                        "is_day": 1,
                        "condition": {
                            "text": "Cloudy",
                            "icon": "//cdn.weatherapi.com/weather/64x64/day/119.png",
                            "code": 1006,
                        },
                        "wind_mph": 15.1,
                        "wind_kph": 24.4,
                        "wind_degree": 136,
                        "wind_dir": "SE",
                        "pressure_mb": 1024.0,
                        "pressure_in": 30.7,
                        "precip_mm": 0.0,
                        "precip_in": 0.0,
                        "humidity": 93,
                        "cloud": 80,
                        "feelslike_c": 3.6,
                        "feelslike_f": 38.4,
                        "windchill_c": 3.6,
                        "windchill_f": 38.4,
                        "heatindex_c": 7.3,
                        "heatindex_f": 45.1,
                        "dewpoint_c": 6.2,
                        "dewpoint_f": 43.2,
                        "will_it_rain": 0,
                        "chance_of_rain": "2",
                        "will_it_snow": 0,
                        "chance_of_snow": "0",
                        "vis_km": 10.0,
                        "vis_miles": 6.0,
                        "gust_mph": 26.5,
                        "gust_kph": 42.6,
                    },
                    {
                        "time_epoch": 1616000400,
                        "time": "2021-03-17 17:00",
                        "temp_c": 7.2,
                        "temp_f": 45.0,
                        "is_day": 1,
                        "condition": {
                            "text": "Patchy rain possible",
                            "icon": "//cdn.weatherapi.com/weather/64x64/day/176.png",
                            "code": 1063,
                        },
                        "wind_mph": 15.7,
                        "wind_kph": 25.3,
                        "wind_degree": 137,
                        "wind_dir": "SE",
                        "pressure_mb": 1024.0,
                        "pressure_in": 30.7,
                        "precip_mm": 0.03,
                        "precip_in": 0.0,
                        "humidity": 92,
                        "cloud": 84,
                        "feelslike_c": 3.3,
                        "feelslike_f": 38.0,
                        "windchill_c": 3.3,
                        "windchill_f": 38.0,
                        "heatindex_c": 7.2,
                        "heatindex_f": 45.0,
                        "dewpoint_c": 6.1,
                        "dewpoint_f": 42.9,
                        "will_it_rain": 0,
                        "chance_of_rain": "21",
                        "will_it_snow": 0,
                        "chance_of_snow": "0",
                        "vis_km": 10.0,
                        "vis_miles": 6.0,
                        "gust_mph": 27.4,
                        "gust_kph": 44.2,
                    },
                    {
                        "time_epoch": 1616004000,
                        "time": "2021-03-17 18:00",
                        "temp_c": 7.1,
                        "temp_f": 44.8,
                        "is_day": 1,
                        "condition": {
                            "text": "Patchy rain possible",
                            "icon": "//cdn.weatherapi.com/weather/64x64/day/176.png",
                            "code": 1063,
                        },
                        "wind_mph": 16.3,
                        "wind_kph": 26.3,
                        "wind_degree": 138,
                        "wind_dir": "SE",
                        "pressure_mb": 1024.0,
                        "pressure_in": 30.7,
                        "precip_mm": 0.04,
                        "precip_in": 0.0,
                        "humidity": 92,
                        "cloud": 88,
                        "feelslike_c": 3.1,
                        "feelslike_f": 37.6,
                        "windchill_c": 3.1,
                        "windchill_f": 37.6,
                        "heatindex_c": 7.1,
                        "heatindex_f": 44.8,
                        "dewpoint_c": 5.9,
                        "dewpoint_f": 42.6,
                        "will_it_rain": 0,
                        "chance_of_rain": "0",
                        "will_it_snow": 0,
                        "chance_of_snow": "0",
                        "vis_km": 10.0,
                        "vis_miles": 6.0,
                        "gust_mph": 28.4,
                        "gust_kph": 45.7,
                    },
                    {
                        "time_epoch": 1616007600,
                        "time": "2021-03-17 19:00",
                        "temp_c": 7.0,
                        "temp_f": 44.7,
                        "is_day": 1,
                        "condition": {
                            "text": "Patchy rain possible",
                            "icon": "//cdn.weatherapi.com/weather/64x64/day/176.png",
                            "code": 1063,
                        },
                        "wind_mph": 15.8,
                        "wind_kph": 25.4,
                        "wind_degree": 138,
                        "wind_dir": "SE",
                        "pressure_mb": 1024.0,
                        "pressure_in": 30.7,
                        "precip_mm": 0.02,
                        "precip_in": 0.0,
                        "humidity": 93,
                        "cloud": 90,
                        "feelslike_c": 3.1,
                        "feelslike_f": 37.6,
                        "windchill_c": 3.1,
                        "windchill_f": 37.6,
                        "heatindex_c": 7.0,
                        "heatindex_f": 44.7,
                        "dewpoint_c": 5.9,
                        "dewpoint_f": 42.6,
                        "will_it_rain": 0,
                        "chance_of_rain": "21",
                        "will_it_snow": 0,
                        "chance_of_snow": "0",
                        "vis_km": 10.0,
                        "vis_miles": 6.0,
                        "gust_mph": 27.8,
                        "gust_kph": 44.8,
                    },
                    {
                        "time_epoch": 1616011200,
                        "time": "2021-03-17 20:00",
                        "temp_c": 7.0,
                        "temp_f": 44.5,
                        "is_day": 0,
                        "condition": {
                            "text": "Overcast",
                            "icon": "//cdn.weatherapi.com/weather/64x64/night/122.png",
                            "code": 1009,
                        },
                        "wind_mph": 15.3,
                        "wind_kph": 24.6,
                        "wind_degree": 138,
                        "wind_dir": "SE",
                        "pressure_mb": 1025.0,
                        "pressure_in": 30.7,
                        "precip_mm": 0.0,
                        "precip_in": 0.0,
                        "humidity": 93,
                        "cloud": 92,
                        "feelslike_c": 3.1,
                        "feelslike_f": 37.6,
                        "windchill_c": 3.1,
                        "windchill_f": 37.6,
                        "heatindex_c": 7.0,
                        "heatindex_f": 44.5,
                        "dewpoint_c": 5.9,
                        "dewpoint_f": 42.6,
                        "will_it_rain": 0,
                        "chance_of_rain": "3",
                        "will_it_snow": 0,
                        "chance_of_snow": "0",
                        "vis_km": 10.0,
                        "vis_miles": 6.0,
                        "gust_mph": 27.2,
                        "gust_kph": 43.8,
                    },
                    {
                        "time_epoch": 1616014800,
                        "time": "2021-03-17 21:00",
                        "temp_c": 6.9,
                        "temp_f": 44.4,
                        "is_day": 0,
                        "condition": {
                            "text": "Overcast",
                            "icon": "//cdn.weatherapi.com/weather/64x64/night/122.png",
                            "code": 1009,
                        },
                        "wind_mph": 14.8,
                        "wind_kph": 23.8,
                        "wind_degree": 138,
                        "wind_dir": "SE",
                        "pressure_mb": 1025.0,
                        "pressure_in": 30.8,
                        "precip_mm": 0.0,
                        "precip_in": 0.0,
                        "humidity": 93,
                        "cloud": 94,
                        "feelslike_c": 3.1,
                        "feelslike_f": 37.6,
                        "windchill_c": 3.1,
                        "windchill_f": 37.6,
                        "heatindex_c": 6.9,
                        "heatindex_f": 44.4,
                        "dewpoint_c": 5.9,
                        "dewpoint_f": 42.6,
                        "will_it_rain": 0,
                        "chance_of_rain": "0",
                        "will_it_snow": 0,
                        "chance_of_snow": "0",
                        "vis_km": 10.0,
                        "vis_miles": 6.0,
                        "gust_mph": 26.6,
                        "gust_kph": 42.8,
                    },
                    {
                        "time_epoch": 1616018400,
                        "time": "2021-03-17 22:00",
                        "temp_c": 6.8,
                        "temp_f": 44.3,
                        "is_day": 0,
                        "condition": {
                            "text": "Overcast",
                            "icon": "//cdn.weatherapi.com/weather/64x64/night/122.png",
                            "code": 1009,
                        },
                        "wind_mph": 14.5,
                        "wind_kph": 23.3,
                        "wind_degree": 140,
                        "wind_dir": "SE",
                        "pressure_mb": 1026.0,
                        "pressure_in": 30.8,
                        "precip_mm": 0.0,
                        "precip_in": 0.0,
                        "humidity": 93,
                        "cloud": 94,
                        "feelslike_c": 3.1,
                        "feelslike_f": 37.5,
                        "windchill_c": 3.1,
                        "windchill_f": 37.5,
                        "heatindex_c": 6.8,
                        "heatindex_f": 44.3,
                        "dewpoint_c": 5.8,
                        "dewpoint_f": 42.5,
                        "will_it_rain": 0,
                        "chance_of_rain": "3",
                        "will_it_snow": 0,
                        "chance_of_snow": "0",
                        "vis_km": 10.0,
                        "vis_miles": 6.0,
                        "gust_mph": 26.1,
                        "gust_kph": 42.0,
                    },
                    {
                        "time_epoch": 1616022000,
                        "time": "2021-03-17 23:00",
                        "temp_c": 6.8,
                        "temp_f": 44.2,
                        "is_day": 0,
                        "condition": {
                            "text": "Overcast",
                            "icon": "//cdn.weatherapi.com/weather/64x64/night/122.png",
                            "code": 1009,
                        },
                        "wind_mph": 14.2,
                        "wind_kph": 22.8,
                        "wind_degree": 142,
                        "wind_dir": "SE",
                        "pressure_mb": 1026.0,
                        "pressure_in": 30.8,
                        "precip_mm": 0.0,
                        "precip_in": 0.0,
                        "humidity": 93,
                        "cloud": 94,
                        "feelslike_c": 3.0,
                        "feelslike_f": 37.5,
                        "windchill_c": 3.0,
                        "windchill_f": 37.5,
                        "heatindex_c": 6.8,
                        "heatindex_f": 44.2,
                        "dewpoint_c": 5.8,
                        "dewpoint_f": 42.4,
                        "will_it_rain": 0,
                        "chance_of_rain": "3",
                        "will_it_snow": 0,
                        "chance_of_snow": "0",
                        "vis_km": 10.0,
                        "vis_miles": 6.0,
                        "gust_mph": 25.6,
                        "gust_kph": 41.2,
                    },
                ],
            },
        ],
    },
}



import json
from typing import Any
from typing import cast
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import apsw
import sqlalchemy.types
from shillelagh.adapters.base import Adapter
from shillelagh.backends.apsw import db
from shillelagh.backends.apsw.vt import VTTable
from shillelagh.exceptions import ProgrammingError
from sqlalchemy import exc
from sqlalchemy.dialects.sqlite.base import SQLiteDialect
from sqlalchemy.engine.url import URL
from sqlalchemy.exc import NoSuchTableError
from sqlalchemy.pool.base import _ConnectionFairy
from sqlalchemy.sql.visitors import VisitableType
from typing_extensions import TypedDict


class SQLAlchemyColumn(TypedDict):
    name: str
    type: VisitableType
    nullable: bool
    default: Optional[str]
    autoincrement: str
    primary_key: int


class APSWDialect(SQLiteDialect):
    name = "shillelagh"
    driver = "apsw"

    @classmethod
    def dbapi(cls):
        return db

    def __init__(
        self,
        adapters: Optional[List[str]] = None,
        adapter_args: Optional[Dict[str, Any]] = None,
        safe: bool = False,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self._adapters = adapters
        self._adapter_args = adapter_args
        self._safe = safe

    def create_connect_args(
        self,
        url: URL,
    ) -> Tuple[
        Tuple[str, Optional[List[str]], Optional[Dict[str, Any]], bool, Optional[str]],
        Dict[str, Any],
    ]:
        path = str(url.database) if url.database else ":memory:"
        return (
            (
                path,
                self._adapters,
                self._adapter_args,
                self._safe,
                self.isolation_level,
            ),
            {},
        )

    def do_ping(self, dbapi_connection: _ConnectionFairy) -> bool:
        return True

    # needed for SQLAlchemy
    def _get_table_sql(
        self,
        connection: _ConnectionFairy,
        table_name: str,
        schema: Optional[str] = None,
        **kwargs: Any,
    ) -> str:
        adapter = self._get_adapter_for_table_name(connection, table_name)
        table = VTTable(adapter)
        return table.get_create_table(table_name)

    def get_columns(
        self,
        connection: _ConnectionFairy,
        table_name: str,
        schema: Optional[str] = None,
        **kwargs: Any,
    ) -> List[SQLAlchemyColumn]:
        adapter = self._get_adapter_for_table_name(connection, table_name)
        columns = adapter.get_columns()
        return [
            {
                "name": column_name,
                "type": getattr(sqlalchemy.types, field.type),
                "nullable": True,
                "default": None,
                "autoincrement": "auto",
                "primary_key": 0,
            }
            for column_name, field in columns.items()
        ]

    def _get_adapter_for_table_name(
        self,
        connection: _ConnectionFairy,
        table_name: str,
    ) -> Adapter:
        raw_connection = cast(db.Connection, connection.engine.raw_connection())
        for adapter in raw_connection._adapters:
            if adapter.supports(table_name):
                break
        else:
            raise ProgrammingError(f"Unsupported table: {table_name}")

        uri_args = adapter.parse_uri(table_name)
        adapter_args = raw_connection._adapter_args.get(adapter.__name__.lower(), ())
        return adapter(*uri_args, *adapter_args)


class APSWGSheetsDialect(APSWDialect):
    """Drop-in replacement for gsheetsdb."""

    name = "gsheets"

    def __init__(
        self,
        service_account_file: Optional[str] = None,
        service_account_info: Optional[Dict[str, Any]] = None,
        subject: Optional[str] = None,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)

        self.service_account_info = service_account_info
        if service_account_file:
            with open(service_account_file) as fp:
                self.service_account_info = json.load(fp)
        self.subject = subject

    def create_connect_args(
        self,
        url: URL,
    ) -> Tuple[
        Tuple[str, Optional[List[str]], Optional[Dict[str, Any]], bool, Optional[str]],
        Dict[str, Any],
    ]:
        adapter_args: Dict[str, Any] = {}
        if self.service_account_info:
            adapter_args["gsheetsapi"] = (self.service_account_info, self.subject)

        return (
            ":memory:",
            ["gsheetsapi"],
            adapter_args,
            True,
            self.isolation_level,
        ), {}

    def get_schema_names(
        self, connection: _ConnectionFairy, **kwargs: Any
    ) -> List[str]:
        return []


class APSWSafeDialect(APSWDialect):
    def __init__(
        self,
        adapters: Optional[List[str]] = None,
        adapter_args: Optional[Dict[str, Any]] = None,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self._adapters = adapters
        self._adapter_args = adapter_args
        self._safe = True

    def create_connect_args(
        self,
        url: URL,
    ) -> Tuple[
        Tuple[str, Optional[List[str]], Optional[Dict[str, Any]], bool, Optional[str]],
        Dict[str, Any],
    ]:
        return (
            (
                ":memory:",
                self._adapters,
                self._adapter_args,
                True,
                self.isolation_level,
            ),
            {},
        )



import urllib.parse
from datetime import date
from datetime import datetime
from datetime import timedelta
from typing import Dict
from typing import Iterator
from typing import List
from typing import Tuple

import dateutil.parser
import requests
import requests_cache
from shillelagh.adapters.base import Adapter
from shillelagh.fields import Boolean
from shillelagh.fields import DateTime
from shillelagh.fields import Float
from shillelagh.fields import Integer
from shillelagh.fields import Order
from shillelagh.fields import String
from shillelagh.filters import Filter
from shillelagh.filters import Range
from shillelagh.types import RequestedOrder
from shillelagh.types import Row


requests_cache.install_cache(
    cache_name="weatherapi_cache",
    backend="sqlite",
    expire_after=180,
)


class WeatherAPI(Adapter):

    safe = True

    time = DateTime(filters=[Range], order=Order.ASCENDING, exact=False)
    time_epoch = Float(filters=[Range], order=Order.ASCENDING, exact=False)
    temp_c = Float()
    temp_f = Float()
    is_day = Boolean()
    wind_mph = Float()
    wind_kph = Float()
    wind_degree = Integer()
    wind_dir = String()
    pressure_mb = Float()
    pressure_in = Float()
    precip_mm = Float()
    precip_in = Float()
    humidity = Integer()
    cloud = Integer()
    feelslike_c = Float()
    feelslike_f = Float()
    windchill_c = Float()
    windchill_f = Float()
    heatindex_c = Float()
    heatindex_f = Float()
    dewpoint_c = Float()
    dewpoint_f = Float()
    will_it_rain = Boolean()
    chance_of_rain = String()
    will_it_snow = Boolean()
    chance_of_snow = String()
    vis_km = Float()
    vis_miles = Float()
    gust_mph = Float()
    gust_kph = Float()

    @staticmethod
    def supports(uri: str) -> bool:
        """https://api.weatherapi.com/v1/history.json?key=XXX&q=94158"""
        parsed = urllib.parse.urlparse(uri)
        query_string = urllib.parse.parse_qs(parsed.query)
        return (
            parsed.netloc == "api.weatherapi.com"
            and parsed.path == "/v1/history.json"
            and "key" in query_string
            and "q" in query_string
        )

    @staticmethod
    def parse_uri(uri: str) -> Tuple[str, str]:
        parsed = urllib.parse.urlparse(uri)
        query_string = urllib.parse.parse_qs(parsed.query)
        location = query_string["q"][0]
        api_key = query_string["key"][0]

        return (location, api_key)

    def __init__(self, location: str, api_key: str):
        self.location = location
        self.api_key = api_key

    def get_data(
        self,
        bounds: Dict[str, Filter],
        order: List[Tuple[str, RequestedOrder]],
    ) -> Iterator[Row]:
        time_range = bounds.get("time", Range(None, None, False, False))
        if not isinstance(time_range, Range):
            raise Exception("Invalid filter")

        #  convert time_epoch range to datetime so we can combine
        #  with the time range
        time_epoch_range = bounds.get("time_epoch", Range(None, None, False, False))
        if not isinstance(time_epoch_range, Range):
            raise Exception("Invalid filter")
        time_epoch_range.start = (
            datetime.utcfromtimestamp(time_epoch_range.start)
            if time_epoch_range.start is not None
            else None
        )
        time_epoch_range.end = (
            datetime.utcfromtimestamp(time_epoch_range.end)
            if time_epoch_range.end is not None
            else None
        )
        time_range += time_epoch_range

        today = date.today()
        start = (
            time_range.start.date() if time_range.start else today - timedelta(days=7)
        )
        end = time_range.end.date() if time_range.end else today

        while start <= end:
            url = (
                f"https://api.weatherapi.com/v1/history.json?key={self.api_key}"
                f"&q={self.location}&dt={start}"
            )
            response = requests.get(url)
            if response.ok:
                payload = response.json()
                hourly_data = payload["forecast"]["forecastday"][0]["hour"]
                columns = self.get_columns()
                for record in hourly_data:
                    row = {column: record[column] for column in columns}
                    row["time"] = dateutil.parser.parse(record["time"])
                    row["rowid"] = int(row["time_epoch"])
                    yield row

            start += timedelta(days=1)



import inspect
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import TypeVar

from shillelagh.exceptions import ProgrammingError
from shillelagh.fields import Field
from shillelagh.filters import Filter
from shillelagh.types import RequestedOrder
from shillelagh.types import Row


T = TypeVar("T", bound="Adapter")


class Adapter:

    # disable "unsafe" adapters that write to disk
    safe = False

    @staticmethod
    def supports(uri: str) -> bool:
        raise NotImplementedError("Subclasses must implement `supports`")

    @classmethod
    def from_uri(cls: Type[T], uri: str) -> T:
        args = cls.parse_uri(uri)
        return cls(*args)

    @staticmethod
    def parse_uri(uri: str) -> Tuple[Any, ...]:
        raise NotImplementedError("Subclasses must implement `parse_uri`")

    def get_columns(self) -> Dict[str, Field]:
        """This method is called for every query, so make sure it's cheap."""
        return dict(
            inspect.getmembers(self, lambda attribute: isinstance(attribute, Field)),
        )

    def get_data(
        self,
        bounds: Dict[str, Filter],
        order: List[Tuple[str, RequestedOrder]],
    ) -> Iterator[Row]:
        raise NotImplementedError("Subclasses must implement `get_data`")

    def insert_row(self, row: Row) -> int:
        raise NotImplementedError("Subclasses must implement `insert_row`")

    def delete_row(self, row_id: int) -> None:
        raise NotImplementedError("Subclasses must implement `delete_row`")

    def update_row(self, row_id: int, row: Row) -> None:
        # Subclasses are free to implement inplace updates
        self.delete_row(row_id)
        self.insert_row(row)

    def close(self) -> None:
        pass



import re
import urllib.parse
from typing import Any
from typing import Dict
from typing import Iterator
from typing import Optional
from typing import Tuple
from unittest import mock

import apsw
import pytest
from shillelagh.adapters.base import Adapter
from shillelagh.backends.apsw.db import connect
from shillelagh.backends.apsw.db import Connection
from shillelagh.backends.apsw.db import Cursor
from shillelagh.exceptions import NotSupportedError
from shillelagh.exceptions import ProgrammingError
from shillelagh.fields import Float
from shillelagh.fields import Integer
from shillelagh.fields import Order
from shillelagh.fields import String
from shillelagh.filters import Equal
from shillelagh.filters import Filter
from shillelagh.filters import Range
from shillelagh.types import Row

from ...fakes import FakeAdapter
from ...fakes import FakeEntryPoint


def test_connect(mocker):
    entry_points = [FakeEntryPoint("dummy", FakeAdapter)]
    mocker.patch(
        "shillelagh.backends.apsw.db.iter_entry_points",
        return_value=entry_points,
    )

    connection = connect(":memory:", ["dummy"], isolation_level="IMMEDIATE")
    cursor = connection.cursor()

    assert cursor.rowcount == -1

    cursor.execute(
        """INSERT INTO "dummy://" (age, name, pets) VALUES (6, 'Billy', 1)""",
    )
    assert cursor.rowcount == -1
    cursor.execute("""DELETE FROM "dummy://" WHERE name = 'Billy'""")
    assert cursor.rowcount == -1

    cursor.execute('SELECT * FROM "dummy://"')
    assert cursor.fetchall() == [(20.0, "Alice", 0), (23.0, "Bob", 3)]
    assert cursor.rowcount == 2

    cursor.execute('SELECT * FROM "dummy://"')
    assert cursor.fetchone() == (20.0, "Alice", 0)
    assert cursor.rowcount == 2
    assert cursor.fetchone() == (23.0, "Bob", 3)
    assert cursor.rowcount == 2
    assert cursor.fetchone() is None

    cursor.execute('SELECT * FROM "dummy://" WHERE age > 21')
    assert cursor.fetchone() == (23.0, "Bob", 3)
    assert cursor.rowcount == 1
    assert cursor.fetchone() is None

    cursor.execute('SELECT * FROM "dummy://"')
    assert cursor.fetchmany() == [(20.0, "Alice", 0)]
    assert cursor.fetchmany(1000) == [(23.0, "Bob", 3)]
    assert cursor.fetchall() == []
    assert cursor.rowcount == 2


def test_check_closed(mocker):
    connection = connect(":memory:", isolation_level="IMMEDIATE")
    cursor = connection.cursor()

    cursor.close()
    with pytest.raises(ProgrammingError) as excinfo:
        cursor.close()
    assert str(excinfo.value) == "Cursor already closed"

    connection.close()
    with pytest.raises(ProgrammingError) as excinfo:
        connection.close()
    assert str(excinfo.value) == "Connection already closed"


def test_check_result(mocker):
    entry_points = [FakeEntryPoint("dummy", FakeAdapter)]
    mocker.patch(
        "shillelagh.backends.apsw.db.iter_entry_points",
        return_value=entry_points,
    )

    connection = connect(":memory:", ["dummy"], isolation_level="IMMEDIATE")
    cursor = connection.cursor()
    with pytest.raises(ProgrammingError) as excinfo:
        cursor.fetchall()

    assert str(excinfo.value) == "Called before `execute`"


def test_check_invalid_syntax(mocker):
    connection = connect(":memory:", isolation_level="IMMEDIATE")
    with pytest.raises(apsw.SQLError) as excinfo:
        connection.execute("SELLLLECT 1")
    assert str(excinfo.value) == 'SQLError: near "SELLLLECT": syntax error'


def test_unsupported_table(mocker):
    connection = connect(":memory:", isolation_level="IMMEDIATE")
    cursor = connection.cursor()

    with pytest.raises(ProgrammingError) as excinfo:
        cursor.execute('SELECT * FROM "dummy://"')
    assert str(excinfo.value) == "Unsupported table: dummy://"


def test_description(mocker):
    entry_points = [FakeEntryPoint("dummy", FakeAdapter)]
    mocker.patch(
        "shillelagh.backends.apsw.db.iter_entry_points",
        return_value=entry_points,
    )

    connection = connect(":memory:", ["dummy"], isolation_level="IMMEDIATE")
    cursor = connection.cursor()

    assert cursor.description is None
    cursor.execute(
        """INSERT INTO "dummy://" (age, name, pets) VALUES (6, 'Billy', 1)""",
    )
    assert cursor.description is None

    cursor.execute('SELECT * FROM "dummy://"')
    assert cursor.description == [
        ("age", Float, None, None, None, None, True),
        ("name", String, None, None, None, None, True),
        ("pets", Integer, None, None, None, None, True),
    ]


def test_execute_many(mocker):
    entry_points = [FakeEntryPoint("dummy", FakeAdapter)]
    mocker.patch(
        "shillelagh.backends.apsw.db.iter_entry_points",
        return_value=entry_points,
    )

    connection = connect(":memory:", ["dummy"], isolation_level="IMMEDIATE")
    cursor = connection.cursor()

    items = [(6, "Billy", 1), (7, "Timmy", 2)]
    with pytest.raises(NotSupportedError) as excinfo:
        cursor.executemany(
            """INSERT INTO "dummy://" (age, name, pets) VALUES (?, ?, ?)""",
            items,
        )
    assert str(excinfo.value) == "`executemany` is not supported, use `execute` instead"


def test_setsize():
    connection = connect(":memory:", isolation_level="IMMEDIATE")
    cursor = connection.cursor()
    cursor.setinputsizes(100)
    cursor.setoutputsizes(100)


def test_close_connection():
    connection = connect(":memory:", isolation_level="IMMEDIATE")
    cursor1 = connection.cursor()
    cursor2 = connection.cursor()

    cursor1.close()
    cursor1.close = mock.MagicMock()
    cursor2.close = mock.MagicMock()

    connection.close()

    cursor1.close.assert_not_called()
    cursor2.close.assert_called()


def test_transaction(mocker):
    entry_points = [FakeEntryPoint("dummy", FakeAdapter)]
    mocker.patch(
        "shillelagh.backends.apsw.db.iter_entry_points",
        return_value=entry_points,
    )

    connection = connect(":memory:", ["dummy"], isolation_level="IMMEDIATE")

    cursor = connection.cursor()
    cursor._cursor = mock.MagicMock()
    cursor._cursor.execute.side_effect = [
        "",
        apsw.SQLError("SQLError: no such table: dummy://"),
        "",
        "",
        "",
        "",
        "",
        "",
    ]

    assert not cursor.in_transaction
    connection.commit()
    cursor._cursor.execute.assert_not_called()

    cursor.execute('SELECT 1 FROM "dummy://"')
    assert cursor.in_transaction
    cursor._cursor.execute.assert_has_calls(
        [
            mock.call("BEGIN IMMEDIATE"),
            mock.call('SELECT 1 FROM "dummy://"', None),
            mock.call('CREATE VIRTUAL TABLE "dummy://" USING FakeAdapter()'),
            mock.call('SELECT 1 FROM "dummy://"', None),
        ],
    )

    connection.rollback()
    assert not cursor.in_transaction
    cursor.execute("SELECT 2")
    assert cursor.in_transaction
    cursor._cursor.execute.assert_has_calls(
        [
            mock.call("BEGIN IMMEDIATE"),
            mock.call('SELECT 1 FROM "dummy://"', None),
            mock.call('CREATE VIRTUAL TABLE "dummy://" USING FakeAdapter()'),
            mock.call('SELECT 1 FROM "dummy://"', None),
            mock.call("ROLLBACK"),
            mock.call("BEGIN IMMEDIATE"),
            mock.call("SELECT 2", None),
        ],
    )

    connection.commit()
    assert not cursor.in_transaction
    cursor._cursor.execute.assert_has_calls(
        [
            mock.call("BEGIN IMMEDIATE"),
            mock.call('SELECT 1 FROM "dummy://"', None),
            mock.call('CREATE VIRTUAL TABLE "dummy://" USING FakeAdapter()'),
            mock.call('SELECT 1 FROM "dummy://"', None),
            mock.call("ROLLBACK"),
            mock.call("BEGIN IMMEDIATE"),
            mock.call("SELECT 2", None),
            mock.call("COMMIT"),
        ],
    )

    cursor._cursor.reset_mock()
    connection.rollback()
    cursor._cursor.execute.assert_not_called()


def test_connection_context_manager():
    with connect(":memory:", isolation_level="IMMEDIATE") as connection:
        cursor = connection.cursor()
        cursor._cursor = mock.MagicMock()
        cursor.execute("SELECT 2")

    cursor._cursor.execute.assert_has_calls(
        [
            mock.call("BEGIN IMMEDIATE"),
            mock.call("SELECT 2", None),
            mock.call("COMMIT"),
        ],
    )


def test_connect_safe(mocker):
    entry_points = [FakeEntryPoint("dummy", FakeAdapter)]
    mocker.patch(
        "shillelagh.backends.apsw.db.iter_entry_points",
        return_value=entry_points,
    )

    connection = connect(":memory:", ["dummy"], safe=True, isolation_level="IMMEDIATE")
    assert connection._adapters == []



import pytest
from shillelagh.fields import Float
from shillelagh.fields import Integer
from shillelagh.fields import Order
from shillelagh.fields import String
from shillelagh.lib import analyse
from shillelagh.lib import DELETED
from shillelagh.lib import RowIDManager
from shillelagh.lib import update_order


def test_row_id_manager_empty_range():
    with pytest.raises(Exception) as excinfo:
        RowIDManager([])

    assert str(excinfo.value) == "Argument `ranges` cannot be empty"


def test_row_id_manager():
    manager = RowIDManager([range(0, 6)])
    assert list(manager) == [0, 1, 2, 3, 4, 5]

    manager.insert()
    assert list(manager) == [0, 1, 2, 3, 4, 5, 6]

    manager.insert(7)
    assert list(manager) == [0, 1, 2, 3, 4, 5, 6, 7]
    assert manager.ranges == [range(0, 8)]

    manager.insert(9)
    assert list(manager) == [0, 1, 2, 3, 4, 5, 6, 7, 9]
    assert manager.ranges == [range(0, 8), range(9, 10)]

    with pytest.raises(Exception) as excinfo:
        manager.insert(5)
    assert str(excinfo.value) == "Row ID 5 already present"

    manager.delete(9)
    assert list(manager) == [0, 1, 2, 3, 4, 5, 6, 7, -1]
    assert manager.ranges == [range(0, 8), DELETED]

    manager.delete(4)
    assert list(manager) == [0, 1, 2, 3, -1, 5, 6, 7, -1]
    assert manager.ranges == [range(0, 4), DELETED, range(5, 8), DELETED]

    with pytest.raises(Exception) as excinfo:
        manager.delete(9)
    assert str(excinfo.value) == "Row ID 9 not found"

    manager.delete(5)
    assert list(manager) == [0, 1, 2, 3, -1, -1, 6, 7, -1]
    assert manager.ranges == [
        range(0, 4),
        DELETED,
        DELETED,
        range(6, 8),
        DELETED,
    ]

    manager.delete(7)
    assert list(manager) == [0, 1, 2, 3, -1, -1, 6, -1, -1]
    assert manager.ranges == [
        range(0, 4),
        DELETED,
        DELETED,
        range(6, 7),
        DELETED,
        DELETED,
    ]


def test_analyse():
    data = [
        {"int": 1, "float": 10.0, "str": "Alice"},
        {"int": 3, "float": 9.5, "str": "Bob"},
        {"int": 2, "float": 8.0, "str": "Charlie"},
    ]
    num_rows, order, types = analyse(data)
    assert num_rows == 3
    assert order == {
        "int": Order.NONE,
        "float": Order.DESCENDING,
        "str": Order.ASCENDING,
    }
    assert types == {"int": Integer, "float": Float, "str": String}


def test_update_order():
    order = update_order(Order.NONE, previous=None, current=1, num_rows=1)
    assert order == Order.NONE

    order = update_order(order, previous=1, current=2, num_rows=2)
    assert order == Order.ASCENDING

    order = update_order(order, previous=2, current=2, num_rows=3)
    assert order == Order.ASCENDING

    order = update_order(order, previous=2, current=1, num_rows=4)
    assert order == Order.NONE



from typing import Any
from typing import Dict
from typing import Iterator
from typing import Optional
from typing import Tuple

from shillelagh.adapters.base import Adapter
from shillelagh.fields import Float
from shillelagh.fields import Integer
from shillelagh.fields import Order
from shillelagh.fields import String
from shillelagh.filters import Equal
from shillelagh.filters import Filter
from shillelagh.filters import Range
from shillelagh.types import Row

from ..fakes import FakeAdapter


def test_adapter_get_columns():
    adapter = FakeAdapter()
    assert adapter.get_columns() == {
        "age": FakeAdapter.age,
        "name": FakeAdapter.name,
        "pets": FakeAdapter.pets,
    }


def test_adapter_get_data():
    adapter = FakeAdapter()

    data = adapter.get_data({}, [])
    assert list(data) == [
        {"rowid": 0, "name": "Alice", "age": 20, "pets": 0},
        {"rowid": 1, "name": "Bob", "age": 23, "pets": 3},
    ]

    data = adapter.get_data({"name": Equal("Alice")}, [])
    assert list(data) == [
        {"rowid": 0, "name": "Alice", "age": 20, "pets": 0},
    ]

    data = adapter.get_data({"age": Range(20, None, False, False)}, [])
    assert list(data) == [
        {"rowid": 1, "name": "Bob", "age": 23, "pets": 3},
    ]

    data = adapter.get_data({}, [("age", Order.DESCENDING)])
    assert list(data) == [
        {"rowid": 1, "name": "Bob", "age": 23, "pets": 3},
        {"rowid": 0, "name": "Alice", "age": 20, "pets": 0},
    ]


def test_adapter_manipulate_rows():
    adapter = FakeAdapter()

    adapter.insert_row({"rowid": None, "name": "Charlie", "age": 6, "pets": 1})
    data = adapter.get_data({}, [])
    assert list(data) == [
        {"rowid": 0, "name": "Alice", "age": 20, "pets": 0},
        {"rowid": 1, "name": "Bob", "age": 23, "pets": 3},
        {"rowid": 2, "name": "Charlie", "age": 6, "pets": 1},
    ]
    adapter.insert_row({"rowid": 4, "name": "Dani", "age": 40, "pets": 2})
    data = adapter.get_data({}, [])
    assert list(data) == [
        {"rowid": 0, "name": "Alice", "age": 20, "pets": 0},
        {"rowid": 1, "name": "Bob", "age": 23, "pets": 3},
        {"rowid": 2, "name": "Charlie", "age": 6, "pets": 1},
        {"rowid": 4, "name": "Dani", "age": 40, "pets": 2},
    ]

    adapter.delete_row(0)
    data = adapter.get_data({}, [])
    assert list(data) == [
        {"rowid": 1, "name": "Bob", "age": 23, "pets": 3},
        {"rowid": 2, "name": "Charlie", "age": 6, "pets": 1},
        {"rowid": 4, "name": "Dani", "age": 40, "pets": 2},
    ]

    adapter.update_row(1, {"rowid": 1, "name": "Bob", "age": 24, "pets": 4})
    data = adapter.get_data({}, [])
    assert list(data) == [
        {"rowid": 2, "name": "Charlie", "age": 6, "pets": 1},
        {"rowid": 4, "name": "Dani", "age": 40, "pets": 2},
        {"rowid": 1, "name": "Bob", "age": 24, "pets": 4},
    ]


def test_from_uri():
    class FakeAdapter(Adapter):
        @staticmethod
        def parse_uri(uri: str) -> Tuple[str, ...]:
            return tuple(uri.split(":", 1))

        def __init__(self, a: str, b: str):
            self.a = a
            self.b = b

    adapter = FakeAdapter.from_uri("foo:bar")
    assert adapter.a == "foo"
    assert adapter.b == "bar"
